//+------------------------------------------------------------------+
//|                                                OrderIterator.mqh |
//|                                    Copyright 2015, EarnForex.com |
//|                                         http://www.earnforex.com |
//+------------------------------------------------------------------+
#include <OrderMap.mqh>
#include <StatusObject.mqh>
#include <RemainingOrderObject.mqh>
//+------------------------------------------------------------------+
//| Class COrderIterator                                             |
//| Purpose: Iterates through orders recursively to find minimum     |
//|          profit (maximum loss). See JS for detailed explanation. |
//+------------------------------------------------------------------+
#define UNLIMITED -DBL_MAX
#define UNDEFINED DBL_MAX

class COrderIterator
{
public:
   CStatusObject *Status;
   COrderMap *RO; // Remaining orders
   double current_price;
   double unrealized_profit;
   double realized_profit;
   static double min_profit; // Same value across all instances.
   static double max_sell_volume; // Same value across all instances.
   bool skip_above;
   bool skip_below;
   double op_above;
   double op_below;

   COrderIterator(CStatusObject *input_Status, COrderMap &input_RO, double input_current_price, double input_unrealized_profit, double input_realized_profit);
   COrderIterator(void): Status(NULL), RO(NULL), current_price(0), unrealized_profit(0), realized_profit(0) {RO = new COrderMap;}
  ~COrderIterator(void);
  
   void Iterate(double order_price);
   bool CheckSimplicity();
   void CheckPairedOrders();
   double FindPriceAbove();
   double FindPriceBelow();
   void ProcessOrder(double order_price);
   void CalculateMaxUp();
   void CalculateMaxDown();
   void AddExecutedOrderToStatus(CRemainingOrderObject &RO_order);
   void AddSLTPtoRO(CRemainingOrderObject &RO_order);
   void AddBasicSLTPtoRO(double price);
   void RemoveCurrentSLTP();
   void RemoveCurrentSL();
   void RemoveCurrentTP();
   void RemoveOrderFromRO(CRemainingOrderObject &RO_order);
   void RemoveOrderFromStatus(CRemainingOrderObject *RO_order);
   void RemoveSecondExit(double price);
   void CalculateUnrealizedProfit();
   void ActivatePendingOrder(CRemainingOrderObject &RO_order);
};

//+------------------------------------------------------------------+
//| Constructor.                                                     |
//+------------------------------------------------------------------+
COrderIterator::COrderIterator(CStatusObject *input_Status, COrderMap &input_RO, double input_current_price, double input_unrealized_profit, double input_realized_profit)
{
   RO = new COrderMap;
   CDOMObject *order, *new_order;
   
   if (CheckPointer(input_Status) != POINTER_INVALID) Status = new CStatusObject(input_Status.Price(), input_Status.Vol(), input_Status.Type(), input_Status.SL(), input_Status.TP());
   else Status = NULL;

   for (order = input_RO.GetFirstNode(); order != NULL; order = input_RO.GetNextNode())
   {
      new_order = new CRemainingOrderObject(order.Ticket(), order.Price(), order.Vol(), order.Type(), order.SL(), order.TP(), order.Status(), order.Origin(), order.StopLimit());
      RO.Add(new_order);
   }
   
   current_price = input_current_price;
   unrealized_profit = input_unrealized_profit;
   realized_profit = input_realized_profit;
   skip_above = false;
   skip_below = false;
   op_above = UNDEFINED;
   op_below = UNDEFINED;
}


//+------------------------------------------------------------------+
//| Destructor                                                       |
//+------------------------------------------------------------------+
COrderIterator::~COrderIterator(void)
{
   if (CheckPointer(Status) != POINTER_INVALID) delete Status;
   delete RO;
}

// Main recursion function. Processes upper and lower orders and calls itself again.
// skip_above - if true, recursion is not used. Instead, only 'below' branch is called.
void COrderIterator::Iterate(const double order_price)
{
   #ifdef _DEBUG
      Print("Order price: ", order_price, " Current price: ", current_price);
   #endif
	if (order_price != UNDEFINED) ProcessOrder(order_price);
	CalculateUnrealizedProfit();
		
	op_above = FindPriceAbove();
	op_below = FindPriceBelow();
	
   #ifdef _DEBUG
      Print("op_above: ", op_above, " op_below: ", op_below);
   #endif

	bool skip = false;
	skip_below = false;
	// skip_above is set via class property.

	if ((op_above != UNDEFINED) && (op_below != UNDEFINED))
	{
		// Returns true if orders at the immediate price above and below the current price qualify as simple and do not require full recursion to process.
		skip = CheckSimplicity();
      #ifdef _DEBUG
		   Print("skip = ", skip);
		#endif
		// No need to do it if we are already skipping upper half via function call argument.
		if ((!skip) && (!skip_above)) 
		{
			// Check if we are inside of SL/TP of one or more orders and there is a clear minimum profit decision at this point and we won't run recursion.
			CheckPairedOrders();
			#ifdef _DEBUG
			   Print("skip_above: ", skip_above, " skip_below: ", skip_below);
			#endif
		}
	}
	
   #ifdef _DEBUG
   	Print("Current price: ", current_price);
   	Print("skip_above: ", skip_above);
   #endif   	
	if (!skip_above)
	{
		if (op_above != UNDEFINED)
		{
		   COrderIterator *OI_new_a = new COrderIterator(Status, RO, current_price, unrealized_profit, realized_profit);
		   OI_new_a.Iterate(op_above);
		   delete OI_new_a;
		}
		else
		{
			// Calculate Max Up.
			CalculateMaxUp();
		}
	}
	
	if (op_below != UNDEFINED)
	{	
		skip_above = skip;
		if (!skip_below)
		{
		   COrderIterator *OI_new_b = new COrderIterator(Status, RO, current_price, unrealized_profit, realized_profit);
		   OI_new_b.Iterate(op_below);
		   delete OI_new_b;
		}
	}
	else
	{
		// Calculate Max Down only if we do not have an indefinite loss path already. MaxDown will always be worse than indefinite loss from MaxUp.
		if (min_profit != UNLIMITED) CalculateMaxDown();
	}
}

// Checks whether orders above and below are simple and thus qualify for shorter recursion instead of full.
// Orders are considered simple if any of these conditions is true:
// 1. Plain buy/sell without SL or TP.
// 2. An SL or TP of some order, which does not have a paired TP or SL among the orders immediately below the current price.
// 3. Non-plain buy/sell in case the SL/TP does not lie within [OP1; OP2] range.
//    In MT5, there is nuance. Orders have to be of a different type, otherwise, the case is not simple. Checked 2015-03-28.
//	   The same is necessary for MT5 stop-limit orders - stop-limit price should not lie within [OP1; OP2] range. Same type nuance is irrelevant.

// It is enough to check SLs and TPs in the top orders. Bottom SL/TP orders will not contain any pairs for the top ones, if top ones did not have such orders.
bool COrderIterator::CheckSimplicity()
{
	CRemainingOrderObject *RO_order;
	int types_found = 0; // 0 - none, 1 - buy, 2 - sell, 3 - both.
	for (RO_order = RO.GetFirstNodeAtPrice(op_above); (RO_order != NULL) && (RO_order.Price() == op_above); RO_order = RO.GetNextNode())
	{
		// If order is an SL or a TP of some other order, check if its origin has a TP or an SL in the op_below price.
		if (RO_order.Status() == SLTP)
		{
			// Partner SL/TP found - the situation is not simple.
			if ((Status.TP() == op_below) || (Status.SL() == op_below)) return(false);
		}
		else if (RO_order.Status() == Pending)
		{
			if (RO_order.Type() == Buy)
			{
				if (types_found == 0) types_found = 1;
				else if (types_found == 2) types_found = 3;
			}
			else if (RO_order.Type() == Sell)
			{
				if (types_found == 0) types_found = 2;
				else if (types_found == 1) types_found = 3;
			}
			if ((RO_order.SL() != 0) && (RO_order.SL() <= op_above) && (RO_order.SL() >= op_below)) return(false);
			if ((RO_order.TP() != 0) && (RO_order.TP() <= op_above) && (RO_order.TP() >= op_below)) return(false);
		}
		else if (RO_order.Status() == StopLimit)
		{
			if ((RO_order.StopLimit() <= op_above) && (RO_order.StopLimit() >= op_below)) return(false);
		}
	}
	for (RO_order = RO.GetFirstNodeAtPrice(op_below); (RO_order != NULL) && (RO_order.Price() == op_below); RO_order = RO.GetNextNode())
	{
		if (RO_order.Status() == Pending)
		{
			// Not simple, if we have different pending order types on op_below and op_above.
			if (types_found == 3) return(false);
			if ((RO_order.Type() == Buy) && (types_found == 2)) return(false);
			if ((RO_order.Type() == Sell) && (types_found == 1)) return(false);

			if ((RO_order.SL() != 0) && (RO_order.SL() <= op_above) && (RO_order.SL() >= op_below)) return(false);
			if ((RO_order.TP() != 0) && (RO_order.TP() <= op_above) && (RO_order.TP() >= op_below)) return(false);
		}
		else if (RO_order.Status() == StopLimit)
		{
			if ((RO_order.StopLimit() <= op_above) && (RO_order.StopLimit() >= op_below)) return(false);
		}
	}
	
	return(true);
}

// Iterates through op_above orders and checks if all are SLs and TPs that have counterparts and those counterparts are located at op_below.
// Does the same for op_below orders.
// Calculates possible minimum profit for the case of going to op_above and op_below.
// Sets skips depending on where was the worst result.
void COrderIterator::CheckPairedOrders()
{
	CRemainingOrderObject *RO_order;
	bool same_order_sl_tp = false;
	double min_profit_above = 0;
	double min_profit_below = 0;
	skip_above = false;
	skip_below = false;
	for (RO_order = RO.GetFirstNodeAtPrice(op_above); (RO_order != NULL) && (RO_order.Price() == op_above); RO_order = RO.GetNextNode())
	{
		if (RO_order.Status() == SLTP)
		{
			if ((Status.TP() == op_below) || (Status.SL() == op_below))
			{
				same_order_sl_tp = true;
				// Upper order can only be an SL for a Sell.
				if (Status.Type() == Sell)
					min_profit_above += Status.Price() - op_above;
				// Upper order can only be a TP for a Buy.
				else if (Status.Type() == Buy)
					min_profit_above += op_above - Status.Price();
			}
			else return; // SL/TP without a counterpart at op_below is a different case.
		}
		else if (RO_order.Status() == Inactive) return; // Any other order (except Inactive) would make it a much more difficult case.
	}
	// Do the same for lower orders only if upper did not fail.
	if (same_order_sl_tp) for (RO_order = RO.GetFirstNodeAtPrice(op_below); (RO_order != NULL) && (RO_order.Price() == op_below); RO_order = RO.GetNextNode())
	{
		if (RO_order.Status() == SLTP)
		{
			if ((Status.TP() == op_above) || (Status.SL() == op_above))
			{
				// Lower order can only be an SL for a Buy.
				if (Status.Type() == Buy)
					min_profit_below += op_below - Status.Price();
				// Lower order can only be a TP for a Sell.
				else if (Status.Type() == Sell)
					min_profit_below += Status.Price() - op_below;
			}
			else return; // SL/TP without a counterpart at op_above is a different case.
		}
		else if (RO_order.Status() == Inactive) return; // Any other order (except Inactive) would make it a much more difficult case.
	}
	// We can now calculate the optimal direction.
	if (same_order_sl_tp)
	{
		#ifdef _DEBUG
		   Print("min_profit_above = ", min_profit_above, " min_profit_below = ", min_profit_below);
		#endif
		// Do only op_above.
		if (min_profit_above <= min_profit_below) skip_below = true;
		// Do only op_below.
		else skip_above = true;
	}
}

// Is there an order above the current price in the remaining orders.
// Returns either price or undefined.
double COrderIterator::FindPriceAbove()
{
	// Find price which is just above the current price.
	double prev_price = UNDEFINED, prev_valid_price = UNDEFINED;
	bool all_orders_are_inactive = false;
	
	for (CRemainingOrderObject *RO_order = RO.GetFirstNode(); RO_order != NULL; RO_order = RO.GetNextNode())
	{
		if (RO_order.Price() < current_price) break;
		// Check if all the orders on the current price are inactive orders. If so, the price should be ignored.
		if (RO_order.Status() == Inactive)
		{
		   // Switched to true only if it is new price. 
		   // Otherwise remains the same as it was before lest it changes to true when there were non-Inactive orders before on that price.
		   if (prev_price != RO_order.Price()) all_orders_are_inactive = true;
		}
		else
		{
		   // Non-Inactive order resets to false in any case.
		   all_orders_are_inactive = false;
		}
		if (!all_orders_are_inactive) prev_valid_price = RO_order.Price();
		prev_price = RO_order.Price();
	}
	//Print(prev_valid_price);
	return(prev_valid_price);
}

// Is there an order below the current price in the remaining orders.
// Returns either price or undefined.
double COrderIterator::FindPriceBelow()
{
	// Find op which is just below the current price.
	CRemainingOrderObject *RO_order;
	for (RO_order = RO.GetFirstNode(); RO_order != NULL; RO_order = RO.GetNextNode())
	{
		if (RO_order.Price() < current_price)
		{
		   // Find first such price which contains a non-Inactive order.
		   if (RO_order.Status() != Inactive) break;
		}
	}

	// No remaining orders below the current price.
	if ((RO_order == NULL) || (RO_order.Price() > current_price)) return(UNDEFINED);
	// Found an order price below current price.
	else return(RO_order.Price());
}

// Process one or more order, which are located at one closest price level above the current price.
// Actually, it is more of a ProcessPrice().
// In MT5, it is necessary to process orders in certain order (sic!):
// 	First - SL/TP.
//  Second - pending orders in order of ascending ticket - i.e. old first (FIFO). In this simulator, it is order by j ascending.
void COrderIterator::ProcessOrder(double order_price)
{
	// First, do SL/TP if needed.
	if (CheckPointer(Status) != POINTER_INVALID)
	{
		if ((Status.SL() == order_price) || (Status.TP() == order_price))
		{
      	for (CRemainingOrderObject *RO_order = RO.GetFirstNodeAtPrice(order_price); (RO_order != NULL) && (RO_order.Price() == order_price); RO_order = RO.GetNextNode())
      	{
				if (RO_order.Status() == SLTP)
				{
//   Print(Status.SL(), " ", Status.TP());
					RemoveOrderFromStatus(RO_order);
					RemoveOrderFromRO(RO_order);
					break;
				}
			}
		}
	}
	
	// Second, cycle through all the remaining orders at a given price.
	for (CRemainingOrderObject *RO_order = RO.GetFirstNodeAtPrice(order_price); (RO_order != NULL) && (RO_order.Price() == order_price); RO_order = RO.GetNextNode())
	{
		if (RO_order.Status() == Inactive) continue;
		if (RO_order.Status() == StopLimit)
		{
			// Activate relevant pending order.
			ActivatePendingOrder(RO_order);
		}
		else // Pending
		{
			// Open new order, pushing it to Status, removing old SL/TP entries if needed and adding new SL/TP entries to RO if needed.
			// Sorts RO if needed.
  			AddExecutedOrderToStatus(RO_order);
		}
		// Removes order from RO.
		RemoveOrderFromRO(RO_order);
	}
	// Switch current price.
	current_price = order_price;
}

// Calculates the minimum profit (maximum loss) attainable from going all way up with the current orders.
void COrderIterator::CalculateMaxUp()
{
	// Look at the open position (Status) and calculate volume of a sell position. If <= 0 - nothing to do.
	if ((CheckPointer(Status) != POINTER_INVALID) && (Status.Type() == Sell))
	{
		if (Status.Vol() > max_sell_volume) max_sell_volume = Status.Vol();
		unrealized_profit = UNLIMITED;
		min_profit = UNLIMITED;
	}
}

// Calculates the minimum profit (maximum loss) attainable from going all way down with the current orders.
void COrderIterator::CalculateMaxDown()
{
	// Look at the open position (Status) and calculate volume of a buy position. If <= 0 - nothing to do.
	if ((CheckPointer(Status) != POINTER_INVALID) && (Status.Type() == Buy))
	{
		double profit = unrealized_profit + realized_profit - Status.Vol() * current_price;
		if (profit < min_profit) min_profit = profit;
	}
}

// Adds a newly executed order as the Status object and new SL/TP to RO if needed.
void COrderIterator::AddExecutedOrderToStatus(CRemainingOrderObject &RO_order)
{
	// New position.
	if (CheckPointer(Status) == POINTER_INVALID)
	{
      Status = new CStatusObject(RO_order.Price(), RO_order.Vol(), RO_order.Type(), RO_order.SL(), RO_order.TP());
		AddSLTPtoRO(RO_order);
	}
	// Same type - update SL/TP, price, and volume.
	else if (Status.Type() == RO_order.Type())
	{
		if (Status.SL() != RO_order.SL())
		{
			RemoveCurrentSL();
			Status.SetSL(RO_order.SL());
			if (Status.SL() != 0) AddBasicSLTPtoRO(Status.SL());
		}
		if (Status.TP() != RO_order.TP())	
		{
			RemoveCurrentTP();
			Status.SetTP(RO_order.TP());
			if (Status.TP() != 0) AddBasicSLTPtoRO(Status.TP());
		}
		// Calculate price based on volume weight. Also, do a smart rounding to 4th decimal place.
		Status.SetPrice(((Status.Vol() * Status.Price()) + (RO_order.Vol() * RO_order.Price())) / (Status.Vol() + RO_order.Vol()));
		Status.SetVolume(Status.Vol() + RO_order.Vol());
	}
	// Types are different.
	else
	{
		// Calculate realized profit.
		if (Status.Type() == Buy) realized_profit += MathMin(RO_order.Vol(), Status.Vol()) * (RO_order.Price() - Status.Price());
		else/* if (Status.Type() == Sell) */ realized_profit += MathMin(RO_order.Vol(), Status.Vol()) * (Status.Price() - RO_order.Price() - spread);
		
		// Equal volumes. Position is closed.
		if (RO_order.Vol() == Status.Vol())
		{
			//First, need to delete old SL/TP if any.
			RemoveCurrentSLTP();
			delete Status;
		}
		else
		{
			// Existing position will absorb new order.
			if (RO_order.Vol() < Status.Vol())
			{
				Status.SetVolume((Status.Vol() - RO_order.Vol()));
			}
			// New order will absorb existing position.
			else
			{
				Status.SetType(RO_order.Type());
				Status.SetVolume(RO_order.Vol() - Status.Vol());
				//First, need to delete old SL/TP if any.
				RemoveCurrentSLTP();
				Status.SetSL(RO_order.SL());
				Status.SetTP(RO_order.TP());
				AddSLTPtoRO(RO_order);
				Status.SetPrice(RO_order.Price());
			}
		}
	}
}

// Add stop-loss and take-profit to the Remaining Orders array.
void COrderIterator::AddSLTPtoRO(CRemainingOrderObject &RO_order)
{
	// Stop-loss.
	if (RO_order.SL() != 0)
	{
		AddBasicSLTPtoRO(RO_order.SL());
	}
	// Take-profit.
	if (RO_order.TP() != 0)
	{
		AddBasicSLTPtoRO(RO_order.TP());
	}
}

// Add stop-loss or take-profit to the Remaining Orders array.
void COrderIterator::AddBasicSLTPtoRO(double price)
{
	CRemainingOrderObject *RO_new_order = new CRemainingOrderObject(0, price, 0, 0, 0, 0, SLTP, 0, 0);
   RO.Add(RO_new_order);
}

// Remove an order from the Remaining Orders array.
void COrderIterator::RemoveOrderFromRO(CRemainingOrderObject &RO_order)
{
	RO.Delete(RO.IndexOf((CDOMObject*)RO_order));
}

// Remove existing SL/TP (if they exist) of the current position.
void COrderIterator::RemoveCurrentSLTP()
{
	RemoveCurrentSL();
	RemoveCurrentTP();
}

// Remove existing SL of the current position.
void COrderIterator::RemoveCurrentSL()
{
	// At this point all existing "sltp" orders in RO are SL/TP of the current position.
	if (Status.SL() != 0)
	{
      for (CRemainingOrderObject *RO_order = RO.GetFirstNodeAtPrice(Status.SL()); (RO_order != NULL) && (RO_order.Price() == Status.SL()); RO_order = RO.GetNextNode())
      {
         if (RO_order.Status() == SLTP)
         {
            RO.DeleteCurrent();
            return;
         }
      }
   }
}

// Remove existing SL of the current position.
void COrderIterator::RemoveCurrentTP()
{
	// At this point all existing "sltp" orders in RO are SL/TP of the current position.
	if (Status.TP() != 0)
	{
      for (CRemainingOrderObject *RO_order = RO.GetFirstNodeAtPrice(Status.TP()); (RO_order != NULL) && (RO_order.Price() == Status.TP()); RO_order = RO.GetNextNode())
      {
         if (RO_order.Status() == SLTP)
         {
            RO.DeleteCurrent();
            return;
         }
      }
	}
}

// Remove an RO_order's origin from the Status array and Second exit from RO (if necessary).
void COrderIterator::RemoveOrderFromStatus(CRemainingOrderObject *RO_order)
{
   RemoveSecondExit(RO_order.Price());
//Print(EnumToString(CheckPointer(RO_order)));
	if (Status.Type() == Buy) realized_profit += Status.Vol() * (RO_order.Price() - Status.Price());
	else realized_profit += Status.Vol() * (Status.Price() - RO_order.Price() - spread);
//	else/* if (Status.Type() == Sell) */ realized_profit += Status.Vol() * (Status.Price() - RO_order.Price() - spread);
	
	// Delete Status object.
	delete Status;
}

// Remove second exit from RO if required.
void COrderIterator::RemoveSecondExit(double price)
{
   //Print(Status.SL(), " ", Status.TP());
	// Do not continue if either SL or TP is absent.
	if ((Status.SL() == 0) || (Status.TP() == 0)) return;
	
	// We hit the order's SL or TP, now the remaining one should be canceled too.
	double op2;
	if (Status.SL() != price)
	{
		op2 = Status.SL();
	}
	else /*if (Status.TP() != price)*/
	{
		op2 = Status.TP();
	}
	
   for (CRemainingOrderObject *RO_order = RO.GetFirstNodeAtPrice(op2); (RO_order != NULL) && (RO_order.Price() == op2); RO_order = RO.GetNextNode())
   {
      if (RO_order.Status() == SLTP)
      {
         RO.DeleteCurrent();
         return;
      }
   }
}

// Calculate unrealized profit based on Status array.
void COrderIterator::CalculateUnrealizedProfit()
{
	unrealized_profit = 0;
	
	if (CheckPointer(Status) != POINTER_INVALID)
	{
		if (Status.Type() == Buy) unrealized_profit += Status.Vol() * (current_price - Status.Price());
		else /*if (Status.Type() == Sell)*/ unrealized_profit += Status.Vol() * (Status.Price() - current_price - spread);
	}
	
	#ifdef _DEBUG
	   Print("unrealized_profit = ", unrealized_profit, " realized_profit = ", realized_profit);
	#endif
	
	// Check if we have a new max loss.
	double profit = unrealized_profit + realized_profit;
	if (profit < min_profit) min_profit = profit;
}

// Receives price and number of the origin stop-limit order.
void COrderIterator::ActivatePendingOrder(CRemainingOrderObject &RO_order)
{
	// Find it first.
   for (CRemainingOrderObject *RO_inactive_order = RO.GetFirstNodeAtPrice(RO_order.StopLimit()); (RO_inactive_order != NULL) && (RO_inactive_order.Price() == RO_order.StopLimit()); RO_inactive_order = RO.GetNextNode())
   {
      if (RO_inactive_order.Origin() == RO_order.Ticket())
      {
         RO_inactive_order.SetStatus(Pending);
         return;
      }
   }
}
//+------------------------------------------------------------------+