//+------------------------------------------------------------------+
//|                                               RiskCalculator.mq5 |
//|                             Copyright  2014-2015, Andriy Moraru |
//+------------------------------------------------------------------+
#property copyright "Copyright  2014-2015, Andriy Moraru, EarnForex.com"
#property link      "http://www.earnforex.com"
#property version   "1.01"
#property indicator_separate_window
#property strict

#property description "Calculates total risk based on the existing positions and pending orders."
#property description "Known issues:"
#property description "1. The results may be somewhat inaccurate when quote currency is different from account currency."
#property description "2. Too many pending orders with SL/TP may be too slow to process."
#property description "3. Assumes constant spread on each tick."
#property description "4. Ignores lack of margin for order execution."
#property description "5. Ignores correlations of currency pairs."
#property description "6. Ignores triangular and other forms of arbitrage."
#property description "7. Does not take into account price slippage."

#include <OrderMap.mqh>
#include <StatusObject.mqh>
#include <RemainingOrderObject.mqh>
#include <OrderIterator.mqh>

// Main
input bool   CalculateSpreads = true; // CalculateSpreads - If true, potential loss due to spreads will become the part of the potential maximum loss.
input bool   CalculateSwaps = true; // CalculateSwaps - If true, accrued swaps will become the part of the potential maximum loss.
input double CommissionPerLot = 0; // CommissionPerLot - Commission charged per lot (one side) in account currency.
input bool   UseEquityInsteadOfBalance = false; // UseEquityInsteadOfBalance - If true, Account Equity will be used instead of Account Balance.
// Font
input color  cpFontColor = clrAzure; // cpFontColor - Font color to ouput the currency pair names.
input color  mnFontColor = clrPaleGoldenrod; // mnFontColor - Font color to ouput the risk in money form.
input color  pcFontColor = clrLimeGreen; // pcFontColor - Font color to output the risk in percentage form.
input string FontFace  = "Courier"; // FontFace - Font name.
input int    FontSize  = 8; // FontSize - Font size.
// Spacing
input int    scaleY = 15; // scaleY - Number of pixels per line in output.
input int    offsetX = 20; // offsetX - Horizontal offset for output.
input int    offsetY = 20; // offsetY - Vertical offset for output.

// Main object for calculating minimum profit (maximum loss) with its static variables initialized.
double COrderIterator::min_profit = UNDEFINED;
double COrderIterator::max_sell_volume = 0;
COrderIterator *OrderIterator;

// Global variable for the current output line's vertical indent.
int Y;
// Global variable for the separate indicator window number;
int Window = -1;

// Global variable to store loss due to swaps of the currently open positions. In account currency.
double swap;
// Global variable for spread - it is used in OrderIterator instances but is defined for each currency pair in this file.
double spread;
// Global variable to store loss due to commission that will be incurred based on current trades. In account currency.
double commission;
// Global variable to store the detected account currency.
string AccCurrency;
// Global variable to store the calculated total risk in money form.
double total_risk;

#ifdef _DEBUG   
   bool single_run = false;
#endif
 
//+------------------------------------------------------------------+
//| Initialization function.                                         |
//+------------------------------------------------------------------+
void OnInit()
{
   IndicatorSetString(INDICATOR_SHORTNAME, "Risk Calculator");
}

//+------------------------------------------------------------------+
//| Deinitialization function.                                       |
//+------------------------------------------------------------------+
void OnDeinit(const int reason)
{
   ObjectsDeleteAll(0, Window, OBJ_LABEL);
   ChartRedraw();
}

//+------------------------------------------------------------------+
//| Indicator calculation event.                                     |
//+------------------------------------------------------------------+
int OnCalculate(const int rates_total,      // size of the price[] array
                 const int prev_calculated,  // bars handled on a previous call
                 const int begin,            // where the significant data start from
                 const double& price[]       // array to calculate
)
{
   // If could not find account currency, probably not connected.
   AccCurrency = AccountInfoString(ACCOUNT_CURRENCY);
   if (AccCurrency == "") return(0);

   CalculateRisk();
   
   return(0);
}

//+------------------------------------------------------------------+
//| Trade event handler.                                             |
//+------------------------------------------------------------------+
void OnTrade()
{
   CalculateRisk();
}

//+------------------------------------------------------------------+
//| Main calculation function. Called from event handlers.           |
//+------------------------------------------------------------------+
void CalculateRisk()
{
   int total;
   #ifdef _DEBUG   
      if (single_run) return;
   #endif 
   
   if (Window == -1) Window = ChartWindowFind();
   ObjectsDeleteAll(0, Window, OBJ_LABEL);

   Y = 0;
   total_risk = 0;
   
   // Orders.
   total = OrdersTotal();
   for (int i = 0; i < total; i++)
   {
      if (!OrderSelect(OrderGetTicket(i))) continue;
      CheckCurrencyPair(OrderGetString(ORDER_SYMBOL));
   }
   
   // Positions.
   total = PositionsTotal();
   for (int i = 0; i < total; i++)
   {
      if (!PositionSelect(PositionGetSymbol(i))) continue;
      CheckCurrencyPair(PositionGetString(POSITION_SYMBOL));
   }

   OutputTotalRisk(total_risk);
   
   #ifdef _DEBUG   
      single_run = true;
   #endif
   ChartRedraw();
}

//+-----------------------------------------------------------------------------------------+
//| Checks if currency pair should be processed and adjusts total_risk calculation function.|
//+-----------------------------------------------------------------------------------------+
void CheckCurrencyPair(string cp)
{
   // This currency pair has already been processed - there is a TEXT_LABEL with its name.
   if (ObjectFind(0, cp) >= 0) return;

   // Zero swap loss before proceeding to the next currency pair.
   swap = 0;
   commission = 0;
   
   double risk = ProcessCurrencyPair(cp);
   // UNDEFINED is stronger than UNLIMITED. UNLIMITED is stronger than any number.
   if ((risk != UNLIMITED) && (total_risk != UNLIMITED) && (risk != UNDEFINED) && (total_risk != UNDEFINED)) total_risk += risk;
   else if (risk == UNDEFINED) total_risk = UNDEFINED;
   else if (total_risk != UNDEFINED) total_risk = UNLIMITED;
}

//+------------------------------------------------------------------+
//| Calculates and outputs risk for a given currency pair.           |
//+------------------------------------------------------------------+
double ProcessCurrencyPair(string cp)
{
   int total;
   COrderIterator::min_profit = UNDEFINED;
   COrderIterator::max_sell_volume = 0;
	OrderIterator = new COrderIterator();

   CDOMObject *order;
   
   // Maximum loss (risk) in account currency.
   double MoneyRisk = 0;
   
   if (CalculateSpreads) spread = SymbolInfoInteger(cp, SYMBOL_SPREAD) * SymbolInfoDouble(cp, SYMBOL_POINT);
   else spread = 0;
   
   total = OrdersTotal();
   for (int i = 0; i < total; i++)
   {
      ulong ticket = OrderGetTicket(i);
      if (!OrderSelect(ticket)) continue;

      if (OrderGetString(ORDER_SYMBOL) != cp) continue;
      
      ENUM_ORDER_TYPE ot = (ENUM_ORDER_TYPE)OrderGetInteger(ORDER_TYPE);
      double volume = OrderGetDouble(ORDER_VOLUME_CURRENT);
      double sl = OrderGetDouble(ORDER_SL);
      double tp = OrderGetDouble(ORDER_TP);
      
      commission += CommissionPerLot * volume * 2;
      if ((ot == ORDER_TYPE_SELL_LIMIT) || (ot == ORDER_TYPE_SELL_STOP))
      {
         order = new CRemainingOrderObject(ticket, OrderGetDouble(ORDER_PRICE_OPEN), volume, Sell, sl == 0 ? 0 : sl - spread, tp == 0 ? 0 : tp - spread, Pending, 0, 0);
         OrderIterator.RO.Add(order);
      }
      else if ((ot == ORDER_TYPE_BUY_STOP) || (ot == ORDER_TYPE_BUY_LIMIT))
      {
         order = new CRemainingOrderObject(ticket, OrderGetDouble(ORDER_PRICE_OPEN) - spread, volume, Buy, sl, tp, Pending, 0, 0);
         OrderIterator.RO.Add(order);
      }
      else if (ot == ORDER_TYPE_SELL_STOP_LIMIT)
      {
         double stop_limit = OrderGetDouble(ORDER_PRICE_STOPLIMIT);
         // Add stop-limit order.
         order = new CRemainingOrderObject(ticket, OrderGetDouble(ORDER_PRICE_OPEN), volume, Sell, sl == 0 ? 0 : sl - spread, tp == 0 ? 0 : tp - spread, StopLimit, 0, stop_limit);
         OrderIterator.RO.Add(order);
         // Add inactive limit order.
         order = new CRemainingOrderObject(ticket, stop_limit, volume, Sell, sl == 0 ? 0 : sl - spread, tp == 0 ? 0 : tp - spread, Inactive, ticket, stop_limit);
         OrderIterator.RO.Add(order);
      }
      else if (ot == ORDER_TYPE_BUY_STOP_LIMIT)
      {
         double stop_limit = OrderGetDouble(ORDER_PRICE_STOPLIMIT);
         // Add stop-limit order.
         order = new CRemainingOrderObject(ticket, OrderGetDouble(ORDER_PRICE_OPEN) - spread, volume, Buy, sl, tp, StopLimit, 0, stop_limit - spread);
         OrderIterator.RO.Add(order);
         // Add inactive limit order.
         order = new CRemainingOrderObject(ticket, stop_limit - spread, volume, Buy, sl, tp, Inactive, ticket, stop_limit);
         OrderIterator.RO.Add(order);
      }
   }      
   
   total = PositionsTotal();
   for (int i = 0; i < total; i++)
   {
      if (!PositionSelect(PositionGetSymbol(i))) continue;
      if (PositionGetString(POSITION_SYMBOL) != cp) continue;
      ENUM_POSITION_TYPE pt = (ENUM_POSITION_TYPE)PositionGetInteger(POSITION_TYPE);
      double sl = PositionGetDouble(POSITION_SL);
      double tp = PositionGetDouble(POSITION_TP);
      
      if (CalculateSwaps) swap += PositionGetDouble(POSITION_SWAP);
      double volume = PositionGetDouble(POSITION_VOLUME);
      commission += CommissionPerLot * volume;
      if (pt == POSITION_TYPE_BUY)
      {
         OrderIterator.Status = new CStatusObject(PositionGetDouble(POSITION_PRICE_OPEN), volume, Buy, sl, tp);
         if (sl)
         {
            order = new CRemainingOrderObject(0, sl, 0, 0, 0, 0, SLTP, 0, 0);
            OrderIterator.RO.Add(order);
         }
         if (tp)
         {
            order = new CRemainingOrderObject(0, tp, 0, 0, 0, 0, SLTP, 0, 0);
            OrderIterator.RO.Add(order);
         }
      }
      else if (pt == POSITION_TYPE_SELL)
      {
         OrderIterator.Status = new CStatusObject(PositionGetDouble(POSITION_PRICE_OPEN), volume, Sell, sl == 0 ? 0 : sl - spread, tp == 0 ? 0 : tp - spread);
         if (sl)
         {
            order = new CRemainingOrderObject(0, sl == 0 ? 0 : sl - spread, 0, 0, 0, 0, SLTP, 0, 0);
            OrderIterator.RO.Add(order);
         }
         if (tp)
         {
            order = new CRemainingOrderObject(0, tp == 0 ? 0 : tp - spread, 0, 0, 0, 0, SLTP, 0, 0);
            OrderIterator.RO.Add(order);
         }
      }
      break; // Only one position per currency pair possible.
   }
   
   OrderIterator.current_price = SymbolInfoDouble(cp, SYMBOL_BID);

	OrderIterator.Iterate(UNDEFINED);

   delete OrderIterator;

   MoneyRisk = Output(cp);

   return(MoneyRisk);
}

//+------------------------------------------------------------------+
//| Creates output for one currency pair via a graphical object.     |
//| Returns: risk in account currency.                               |
//+------------------------------------------------------------------+
double Output(string cp)
{
   string RiskOutput = "", SecondRiskOutput = "";
   double MoneyRisk;
   if (COrderIterator::min_profit == UNDEFINED) RiskOutput = JustifyRight("Undefined", 25);
   else if (COrderIterator::min_profit == UNLIMITED) RiskOutput = JustifyRight("Unlimited", 25) +  " (" + DoubleToString(COrderIterator::max_sell_volume, 2) + " lot)";
   else
   {
      double UnitCost = SymbolInfoDouble(cp, SYMBOL_TRADE_TICK_VALUE);
      double OnePoint = SymbolInfoDouble(cp, SYMBOL_POINT);
      MoneyRisk = -COrderIterator::min_profit * UnitCost / OnePoint - commission;
      if (CalculateSwaps) MoneyRisk -= swap;
      double Size;
      if (UseEquityInsteadOfBalance) Size = AccountInfoDouble(ACCOUNT_EQUITY);
      else Size = AccountInfoDouble(ACCOUNT_BALANCE);
      double PercentageRisk = (MoneyRisk / Size) * 100;
      RiskOutput = JustifyRight(FormatNumber(DoubleToString(MoneyRisk, 2)) + " " + AccCurrency, 25);
      SecondRiskOutput = JustifyRight(DoubleToString(PercentageRisk, 2) + "%", 25);
   }
      
   ObjectCreate(0, cp, OBJ_LABEL, Window, 0, 0);
   ObjectSetString(0, cp, OBJPROP_TEXT, cp);
   ObjectSetInteger(0, cp, OBJPROP_FONTSIZE, FontSize);
   ObjectSetString(0, cp, OBJPROP_FONT, FontFace);
   ObjectSetInteger(0, cp, OBJPROP_COLOR, cpFontColor);
   ObjectSetInteger(0, cp, OBJPROP_CORNER, 0);
   ObjectSetInteger(0, cp, OBJPROP_XDISTANCE, offsetX);
   ObjectSetInteger(0, cp, OBJPROP_YDISTANCE, Y * scaleY + offsetY);
   
   ObjectCreate(0, cp + "Amount", OBJ_LABEL, Window, 0, 0);
   ObjectSetString(0, cp + "Amount", OBJPROP_TEXT, RiskOutput);
   ObjectSetInteger(0, cp + "Amount", OBJPROP_FONTSIZE, FontSize);
   ObjectSetString(0, cp + "Amount", OBJPROP_FONT, FontFace);
   ObjectSetInteger(0, cp + "Amount", OBJPROP_COLOR, mnFontColor);
   ObjectSetInteger(0, cp + "Amount", OBJPROP_CORNER, 0);
   ObjectSetInteger(0, cp + "Amount", OBJPROP_XDISTANCE, offsetX + 1 * offsetX);
   ObjectSetInteger(0, cp + "Amount", OBJPROP_YDISTANCE, Y * scaleY + offsetY);

   if (SecondRiskOutput != "")
   {
      ObjectCreate(0, cp + "Percentage", OBJ_LABEL, Window, 0, 0);
      ObjectSetString(0, cp + "Percentage", OBJPROP_TEXT, SecondRiskOutput);
      ObjectSetInteger(0, cp + "Percentage", OBJPROP_FONTSIZE, FontSize);
      ObjectSetString(0, cp + "Percentage", OBJPROP_FONT, FontFace);
      ObjectSetInteger(0, cp + "Percentage", OBJPROP_COLOR, pcFontColor);
      ObjectSetInteger(0, cp + "Percentage", OBJPROP_CORNER, 0);
      ObjectSetInteger(0, cp + "Percentage", OBJPROP_XDISTANCE, offsetX + 6 * offsetX);
      ObjectSetInteger(0, cp + "Percentage", OBJPROP_YDISTANCE, Y * scaleY + offsetY);
   }
   
   Y++;
   
   if ((COrderIterator::min_profit != UNLIMITED) && (COrderIterator::min_profit != UNDEFINED)) return(MoneyRisk);
   else return(COrderIterator::min_profit);
}

//+------------------------------------------------------------------+
//| Creates output for combined risk for all currency pairs.         |
//| Inputs: risk - risk in account currency.                         |
//+------------------------------------------------------------------+
void OutputTotalRisk(double risk)
{
   Y++;
   string RiskOutput = "", SecondRiskOutput = "";;
   if (risk == 0) RiskOutput = JustifyRight("No risk.", 25);
   else if (risk == UNDEFINED) RiskOutput = JustifyRight("Undefined", 25);
   else if (risk == UNLIMITED) RiskOutput = JustifyRight("Unlimited", 25);
   else
   {
      double Size;
      if (UseEquityInsteadOfBalance) Size = AccountInfoDouble(ACCOUNT_EQUITY);
      else Size = AccountInfoDouble(ACCOUNT_BALANCE);
      double PercentageRisk = (risk / Size) * 100;
      //RiskOutput = DoubleToString(risk, 2) + " (" + DoubleToString(PercentageRisk, 2) + "%)";
      RiskOutput = JustifyRight(FormatNumber(DoubleToString(risk, 2)) + " " + AccCurrency, 25);
      SecondRiskOutput = JustifyRight(DoubleToString(PercentageRisk, 2) + "%", 25);
   }
      
   ObjectCreate(0, "Total", OBJ_LABEL, Window, 0, 0);
   ObjectSetString(0, "Total", OBJPROP_TEXT, "Total");
   ObjectSetInteger(0, "Total", OBJPROP_FONTSIZE, FontSize);
   ObjectSetString(0, "Total", OBJPROP_FONT, FontFace);
   ObjectSetInteger(0, "Total", OBJPROP_COLOR, cpFontColor);
   ObjectSetInteger(0, "Total", OBJPROP_CORNER, 0);
   ObjectSetInteger(0, "Total", OBJPROP_XDISTANCE, offsetX);
   ObjectSetInteger(0, "Total", OBJPROP_YDISTANCE, Y * scaleY + offsetY);
   
   ObjectCreate(0, "TotalAmount", OBJ_LABEL, Window, 0, 0);
   ObjectSetString(0, "TotalAmount", OBJPROP_TEXT, RiskOutput);
   ObjectSetInteger(0, "TotalAmount", OBJPROP_FONTSIZE, FontSize);
   ObjectSetString(0, "TotalAmount", OBJPROP_FONT, FontFace);
   ObjectSetInteger(0, "TotalAmount", OBJPROP_COLOR, mnFontColor);
   ObjectSetInteger(0, "TotalAmount", OBJPROP_CORNER, 0);
   ObjectSetInteger(0, "TotalAmount", OBJPROP_XDISTANCE, offsetX + 1 * offsetX);
   ObjectSetInteger(0, "TotalAmount", OBJPROP_YDISTANCE, Y * scaleY + offsetY);

   if ((SecondRiskOutput != "") && (risk != 0))
   {
      ObjectCreate(0, "TotalPercentage", OBJ_LABEL, Window, 0, 0);
      ObjectSetString(0, "TotalPercentage", OBJPROP_TEXT, SecondRiskOutput);
      ObjectSetInteger(0, "TotalPercentage", OBJPROP_FONTSIZE, FontSize);
      ObjectSetString(0, "TotalPercentage", OBJPROP_FONT, FontFace);
      ObjectSetInteger(0, "TotalPercentage", OBJPROP_COLOR, pcFontColor);
      ObjectSetInteger(0, "TotalPercentage", OBJPROP_CORNER, 0);
      ObjectSetInteger(0, "TotalPercentage", OBJPROP_XDISTANCE, offsetX + 6 * offsetX);
      ObjectSetInteger(0, "TotalPercentage", OBJPROP_YDISTANCE, Y * scaleY + offsetY);
   }
}

//+----------------------------------------------------------------------+
//| Returns a number formatted as money string.                          |
//+----------------------------------------------------------------------+
string FormatNumber(const string number)
{
   string output = "";
   int length = StringLen(number);
   int j = 0;
   // Start from first digit from right starting from the decimal separator. E.g. from '4' in '1234.56'.
   for (int i = length - 4; i >= 0; i--)
   {
      if ((j % 3 == 0) && (j != 0)) output = "," + output;
      output = StringSubstr(number, i, 1) + output;
      j++;
   }
   return(output + StringSubstr(number, length - 3, 3));
}

//+----------------------------------------------------------------------+
//| Returns a string with enough spaces added from the left side to make | 
//| it a width length string.                                            |
//+----------------------------------------------------------------------+
string JustifyRight(string text, const int width)
{
   int length = StringLen(text);
   // Cannot do anything - string too long.
   if (length >= width) return(text);
   
   int j = 0;
   // Start from first digit from right starting from the decimal separator. E.g. from '4' in '1234.56'.
   for (int i = width - length; i >= 0; i--)
      text = " " + text;
   
   return(text);
}